import { Vec2 } from "cc";
import GRoleBase from "../../../role/GRoleBase";
import GFSMBase, { GFSMProcessEnum, GFSMProcessInfo, GFSMProcessMode } from "../../GFSMBase";
import { GRoleUtil } from "../../../../entity/GRole";


//流程枚举
enum ProcessEnum {
    //寻找敌人
    SeekEnemy = 0,
    //移动到可攻击范围
    MoveToAttackRange = 1,
    //攻击敌人
    AttackEnemy = 2,
    //回阵型
    MoveToTactical = 3,
}

export default abstract class GFSMBattle extends GFSMBase{

    protected _player:() => GRoleBase<{}>;
    get player():GRoleBase<{}>{
        if(this._player)
            return this._player();
        return null;
    }

    //锁定的敌人
    _enemy:() => GRoleBase<any>;
    get enemy():GRoleBase<{}>{
        if(this._enemy)
            return this._enemy();
        return null;
    }

    //流程图
    process: { [key: number]: GFSMProcessInfo; } = {
        [ProcessEnum.SeekEnemy]:{
            title:"寻找敌人",
            execute: this.onSeekEnemyProcess.bind(this),
            // to:[ProcessEnum.MoveToAttackRange], //移动到可攻击范围
            to:[ProcessEnum.MoveToAttackRange,ProcessEnum.MoveToTactical], //移动到可攻击范围
        },
        [ProcessEnum.MoveToAttackRange]:{
            title:"移动到可攻击范围",
            mode:GFSMProcessMode.WaitExecute,
            execute: this.onMoveToAttackRangeProcess.bind(this),
            to:[ProcessEnum.AttackEnemy] //攻击敌人 回阵型
        },
        [ProcessEnum.AttackEnemy]:{
            title:"攻击敌人",
            mode:GFSMProcessMode.WaitExecute,
            execute: this.onAttackProcess.bind(this),
            to:[ProcessEnum.SeekEnemy],//移动回阵型
        },
        [ProcessEnum.MoveToTactical]:{
            title:"移动回阵型",
            mode:GFSMProcessMode.WaitExecute,
            execute: this.onMoveToTacticalProcess.bind(this),
            to:[ProcessEnum.SeekEnemy],//移动回阵型
        }
    }

    constructor(player:GRoleBase<{}>){
        super();
        this._player = GRoleUtil.get(player);
    }

    abstract onSeekEnemy():GRoleBase<any>;

    //寻敌流程
    onSeekEnemyProcess(dt:number):number{

        //播放等待
        this.player.fsmAnim.isMove = false;
        this.player.fsmAnim.isAttack = false;

        if(this.enemy){
            //如果有敌人 直接 攻击
            return ProcessEnum.MoveToAttackRange;
        }

        if((this._enemy = GRoleUtil.get(this.onSeekEnemy())) && this.enemy){
            //如果有敌人 直接 攻击
            return ProcessEnum.MoveToAttackRange;
        }else{
            //如果没有敌人则回去
            if(!(this.player.tacticalPos.equals(this.player.v2World))){
                return ProcessEnum.MoveToTactical
            }else{
                return GFSMProcessEnum.Reset;
            }
        }

    }

    //移动可攻击范围
    onMoveToAttackRangeProcess(dt:number){

        //播放移动
        this.player.fsmAnim.isMove = true;
        this.player.fsmAnim.isAttack = false;

        //如果没有敌人则重置
        if(!this.enemy) return GFSMProcessEnum.Reset;

        let distance = Vec2.distance(this.player.v2World,this.enemy.v2World);

        //如果在攻击范围则调用攻击 负责 靠近目标
        if(distance < this.player.range){
            return ProcessEnum.AttackEnemy;
        }else{
            //靠近目标
            this.player.onMoveTarget(this.enemy.v2World,dt);
            return GFSMProcessEnum.Wait;
        }

    }

    //移动回阵型
    onMoveToTacticalProcess(dt:number){

        //播放移动
        this.player.fsmAnim.isMove = true;
        this.player.fsmAnim.isAttack = false;

        if(this.player.onMoveTarget(this.player.tacticalPos,dt)){
            this.player.onRecoverTacticalTowards();
            //回到阵容 搜索敌人
            return ProcessEnum.SeekEnemy;
        }

        return GFSMProcessEnum.Wait;

    }
    
    //攻击敌人
    onAttackProcess(dt:number){
        //播放移动
        this.player.fsmAnim.isMove = false;
        this.player.fsmAnim.isAttack = true;
        //如果有敌人则攻击 没有 则 重置
        if(this.enemy){
            return GFSMProcessEnum.Wait;
        }else{
            return ProcessEnum.SeekEnemy;
        }
    }


}

